// 
// Author:
//       smdn <smdn@mail.invisiblefulmoon.net>
// 
// Copyright (c) 2008-2010 smdn
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Collections.Generic;
using System.IO;
using System.Text;

using Smdn.Collections;
using Smdn.IO;
using Smdn.Net.Imap4.Client.Session;

namespace Smdn.Net.Imap4.Client {
  public class ImapMessageInfo : IImapMessageAttribute {
    public ImapOpenedMailboxInfo Mailbox {
      get; private set;
    }

    public long Sequence {
      get; private set;
    }

    public long Uid {
      get; private set;
    }

    public long UidValidity {
      get; private set;
    }

    public Uri Url {
      get { return uriBuilder.Uri; }
    }

    /*
     * IImapMessageDynamicAttribute
     */
    private IImapMessageDynamicAttribute dynamicAttribute;

    internal IImapMessageDynamicAttribute DynamicAttribute {
      get { UpdateDynamicAttributes(false); return dynamicAttribute; }
      set { dynamicAttribute = value; }
    }

    public IImapMessageFlagSet Flags {
      get { return DynamicAttribute.Flags; }
    }

    [CLSCompliant(false)]
    public ulong ModSeq {
      get { return DynamicAttribute.ModSeq; }
    }

    /*
     * IImapMessageStaticAttribute
     */
    private IImapMessageStaticAttribute staticAttribute;

    internal IImapMessageStaticAttribute StaticAttribute {
      get { UpdateStaticAttributes(); return staticAttribute; }
      set { staticAttribute = value; }
    }

    public IImapBodyStructure BodyStructure {
      get { return StaticAttribute.BodyStructure; }
    }

    public ImapEnvelope Envelope {
      get { return StaticAttribute.Envelope; }
    }

    public DateTimeOffset? InternalDate {
      get { return StaticAttribute.InternalDate; }
    }

    public long Length {
      get { return StaticAttribute.Rfc822Size; }
    }

    long IImapMessageStaticAttribute.Rfc822Size {
      get { return StaticAttribute.Rfc822Size; }
    }

    long IImapMessageStaticAttribute.GetBinarySizeOf(string section)
    {
      return StaticAttribute.GetBinarySizeOf(section);
    }

    public bool IsMarkedAsDeleted {
      get { return Flags.Has(ImapMessageFlag.Deleted); }
    }

    private ImapSequenceSet UidSet {
      get
      {
        if (uidSet == null)
          uidSet = ImapSequenceSet.CreateUidSet(Uid);

        return uidSet;
      }
    }

    internal ImapMessageInfo(ImapClient client, ImapOpenedMailboxInfo mailbox, long uid, long uidValidity, long sequence)
    {
      this.client = client;
      this.Mailbox = mailbox;
      this.Uid = uid;
      this.UidValidity = uidValidity;
      this.Sequence = sequence;

      (this as IImapUrl).SetBaseUrl(mailbox.Url);
    }

    private ImapClient client;
    private ImapUriBuilder uriBuilder;
    private ImapSequenceSet uidSet;

    /*
     * operations
     */
    public void Refresh()
    {
      UpdateStaticAttributes();
      UpdateDynamicAttributes(true);
    }

    internal static readonly ImapFetchDataItem FetchStaticDataItem =
      ImapFetchDataItem.BodyStructure +
      ImapFetchDataItem.Envelope + 
      ImapFetchDataItem.InternalDate +
      ImapFetchDataItem.Rfc822Size;

    private void UpdateStaticAttributes()
    {
      if (staticAttribute != null)
        return;

      Mailbox.CheckClosed();

      ImapMessageStaticAttribute[] staticAttr;

      ImapClient.ThrowIfError(client.Session.Fetch(UidSet,
                                                   FetchStaticDataItem,
                                                   out staticAttr));

      if (0 < staticAttr.Length)
        staticAttribute = staticAttr[0];
    }

    internal static ImapFetchDataItem GetFetchDynamicDataItem(ImapCapabilityList serverCapabilities)
    {
      if (serverCapabilities.Has(ImapCapability.CondStore))
        return ImapFetchDataItem.Flags + ImapFetchDataItem.ModSeq;
      else
        return ImapFetchDataItem.Flags;
    }

    private void UpdateDynamicAttributes(bool refresh)
    {
      if (!refresh && dynamicAttribute != null)
        return;

      Mailbox.CheckClosed();

      ImapMessageDynamicAttribute[] dynamicAttr;

      ImapClient.ThrowIfError(client.Session.Fetch(UidSet,
                                                   GetFetchDynamicDataItem(client.ServerCapabilities),
                                                   out dynamicAttr));

      if (0 < dynamicAttr.Length)
        dynamicAttribute = dynamicAttr[0];
    }

    public void AddFlags(ImapMessageFlag flag, params ImapMessageFlag[] flags)
    {
      Store(ImapStoreDataItem.AddFlags(flag, flags));
    }

    public void AddKeywords(string keyword, params string[] keywords)
    {
      Store(ImapStoreDataItem.AddFlags(keyword, keywords));
    }

    public void AddFlags(IImapMessageFlagSet flagsAndKeywords)
    {
      Store(ImapStoreDataItem.AddFlags(flagsAndKeywords));
    }

    public void RemoveFlags(ImapMessageFlag flag, params ImapMessageFlag[] flags)
    {
      Store(ImapStoreDataItem.RemoveFlags(flag, flags));
    }

    public void RemoveKeywords(string keyword, params string[] keywords)
    {
      Store(ImapStoreDataItem.RemoveFlags(keyword, keywords));
    }

    public void RemoveFlags(IImapMessageFlagSet flagsAndKeywords)
    {
      Store(ImapStoreDataItem.RemoveFlags(flagsAndKeywords));
    }

    public void ReplaceFlags(ImapMessageFlag flag, params ImapMessageFlag[] flags)
    {
      Store(ImapStoreDataItem.ReplaceFlags(flag, flags));
    }

    public void ReplaceKeywords(string keyword, params string[] keywords)
    {
      Store(ImapStoreDataItem.ReplaceFlags(keyword, keywords));
    }

    public void ReplaceFlags(IImapMessageFlagSet flagsAndKeywords)
    {
      Store(ImapStoreDataItem.ReplaceFlags(flagsAndKeywords));
    }

    public void ToggleFlags(ImapMessageFlag flag, params ImapMessageFlag[] flags)
    {
      ToggleFlags(new ImapMessageFlagList(flags.Prepend(flag)));
    }

    public void ToggleKeywords(string keyword, params string[] keywords)
    {
      ToggleFlags(new ImapMessageFlagList(keywords.Prepend(keyword)));
    }

    public void ToggleFlags(IImapMessageFlagSet flagsAndKeywords)
    {
      var addFlags = new ImapMessageFlagList();
      var removeFlags = new ImapMessageFlagList();

      foreach (var flag in flagsAndKeywords) {
        if (dynamicAttribute.Flags.Has(flag))
          removeFlags.Add(flag);
        else
          addFlags.Add(flag);
      }

      if (0 < removeFlags.Count)
        Store(ImapStoreDataItem.RemoveFlags(removeFlags));

      if (0 < addFlags.Count)
        Store(ImapStoreDataItem.AddFlags(addFlags));
    }

    public void Store(ImapStoreDataItem storeDataItem)
    {
      if (storeDataItem == null)
        throw new ArgumentNullException("storeDataItem");

      Mailbox.CheckClosed();

      // TODO: update dynamic attributes
      ImapClient.ThrowIfError(client.Session.Store(UidSet, storeDataItem));
    }

    private static ImapStoreDataItem markAsDeletedStoreDataItem = ImapStoreDataItem.AddFlags(ImapMessageFlag.Deleted);

    public void MarkAsDeleted()
    {
      Store(markAsDeletedStoreDataItem);
    }

    public void Delete()
    {
      MarkAsDeleted();

      if (client.ServerCapabilities.Has(ImapCapability.UidPlus))
        ImapClient.ThrowIfError(client.Session.UidExpunge(UidSet));
      else
        ImapClient.ThrowIfError(client.Session.Expunge());
    }

    public void CopyTo(ImapMailboxInfo destinationMailbox)
    {
      if (destinationMailbox == null)
        throw new ArgumentNullException("destinationMailbox");
      if (!destinationMailbox.Exists)
        throw new InvalidOperationException("destination mailbox is not existent");

      Mailbox.CheckClosed();

      ImapClient.ThrowIfError(client.Session.Copy(UidSet, destinationMailbox.Mailbox));
    }

    public void MoveTo(ImapMailboxInfo destinationMailbox)
    {
      CopyTo(destinationMailbox);

      MarkAsDeleted();
    }

    /// <returns>The mailbox if created, otherwise null.</returns>
    public ImapMailboxInfo CopyTo(string destinationMailboxName)
    {
      Mailbox.CheckClosed();

      ImapMailbox createdMailbox;

      ImapClient.ThrowIfError(client.Session.Copy(UidSet,
                                                  ImapClient.GetValidatedMailboxName(destinationMailboxName),
                                                  out createdMailbox));

      if (createdMailbox == null)
        return null;
      else
        return new ImapMailboxInfo(client, createdMailbox);
    }

    /// <returns>The mailbox if created, otherwise null.</returns>
    public ImapMailboxInfo MoveTo(string destinationMailboxName)
    {
      var ret = CopyTo(destinationMailboxName);

      MarkAsDeleted();

      return ret;
    }

    /*
     * Open*(), Read*()
     */
    private static string GetSection(IImapBodyStructure structure)
    {
      if (structure == null)
        throw new ArgumentNullException("structure");

      return structure.Section;
    }

    private static string GetEncoding(IImapBodyStructure structure, bool decode)
    {
      if (decode && structure is ImapSinglePartBodyStructure)
        return (structure as ImapSinglePartBodyStructure).Encoding;
      else
        return null;
    }

    public Stream OpenRead()
    {
      return OpenRead(string.Empty, null, null);
    }

    public Stream OpenRead(bool decode)
    {
      return OpenRead(string.Empty, null, GetEncoding(BodyStructure, decode));
    }

    public Stream OpenRead(IImapBodyStructure section)
    {
      return OpenRead(GetSection(section), null, null);
    }

    public Stream OpenRead(IImapBodyStructure section, bool decode)
    {
      return OpenRead(GetSection(section), null, GetEncoding(BodyStructure, decode));
    }

    public Stream OpenRead(string section)
    {
      return OpenRead(section, null, null);
    }

    public Stream OpenRead(string section, ImapPartialRange range)
    {
      return OpenRead(section, range, null);
    }

    private Stream OpenRead(string section, ImapPartialRange range, string encoding)
    {
      //var decode = (encoding != null);

      Mailbox.CheckClosed();

      var stream = new ImapFetchMessageBodyStream(client.Session,
                                                  true /*TODO: peek*/,
                                                  UidSet,
                                                  section,
                                                  range,
                                                  ImapFetchMessageBodyStream.DefaultFetchBlockSize);

      IImapMessageAttribute discard;

      ImapClient.ThrowIfError(stream.Prepare(null, out discard));

      return stream;
    }

    public TOutput ReadAs<TOutput>(Converter<Stream, TOutput> converter)
    {
      return ReadAsCore(OpenRead(), converter);
    }

    public TOutput ReadAs<TOutput>(bool decode, Converter<Stream, TOutput> converter)
    {
      return ReadAsCore(OpenRead(decode), converter);
    }

    public TOutput ReadAs<TOutput>(IImapBodyStructure section, Converter<Stream, TOutput> converter)
    {
      return ReadAsCore(OpenRead(section), converter);
    }

    public TOutput ReadAs<TOutput>(IImapBodyStructure section, bool decode, Converter<Stream, TOutput> converter)
    {
      return ReadAsCore(OpenRead(section, decode), converter);
    }

    public TOutput ReadAs<TOutput>(string section, Converter<Stream, TOutput> converter)
    {
      return ReadAsCore(OpenRead(section), converter);
    }

    public TOutput ReadAs<TOutput>(string section, ImapPartialRange range, Converter<Stream, TOutput> converter)
    {
      return ReadAsCore(OpenRead(section, range), converter);
    }

    private TOutput ReadAsCore<TOutput>(Stream stream, Converter<Stream, TOutput> converter)
    {
      if (converter == null)
        throw new ArgumentNullException("converter");

      using (stream) {
        return converter(stream);
      }
    }

    public StreamReader OpenText()
    {
      return OpenText(string.Empty, Encoding.Default);
    }

    public StreamReader OpenText(string section)
    {
      return OpenText(section, Encoding.Default);
    }

    public StreamReader OpenText(Encoding encoding)
    {
      return OpenText(string.Empty, encoding);
    }

    public StreamReader OpenText(string section, Encoding encoding)
    {
      if (encoding == null)
        throw new ArgumentNullException("encoding");

      throw new NotImplementedException();
      //return new StreamReader(OpenRead(section, true), encoding);
    }

    public byte[] ReadAllBytes()
    {
      return ReadAllBytes(string.Empty, false);
    }

    public byte[] ReadAllBytes(IImapBodyStructure section)
    {
      return ReadAllBytes(GetSection(section), false);
    }

    public byte[] ReadAllBytes(IImapBodyStructure section, bool decode)
    {
      return ReadAllBytes(GetSection(section), decode);
    }

    public byte[] ReadAllBytes(string section, bool decode)
    {
      using (var stream = OpenRead(section, null, null)) {
        return stream.ReadToEnd();
      }
    }

    public IEnumerable<string> ReadLines(string section)
    {
      throw new NotImplementedException();
    }

    public string[] ReadAllLines(string section)
    {
      return ReadLines(section).ToArray();
    }

    public string ReadAllText(string section)
    {
      throw new NotImplementedException();
    }

    /// <summary>Infrastructure. It is not intended to be used directly from your code.</summary>
    void IImapUrl.SetBaseUrl(Uri baseUrl)
    {
      uriBuilder = new ImapUriBuilder(baseUrl);

      uriBuilder.Uid = Uid;
      uriBuilder.UidValidity = UidValidity;
    }
  }
}
